#!/usr/bin/env python

# Copyright (C) 2006 Christian Dywan <software at twotoasts dot de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import sys

"""Check for and try to include required modules."""
modules_verbose = 0
modules_required = ['os', 'stat', 'time', 'gobject', 'md5',
    'optparse', 'subprocess', 'fnmatch',
    'pygtk=2.0', 'gtk', 'gtk.glade', '?xdg.Mime']
modules_missing = []
modules_skipped = []
for module in modules_required:
    if module[0] == '?':
        module = module[1:]
        module_optional = True
    else:
        module_optional = False
    if module.find('=') <> -1:
        module, module_version = module.split('=')
    else:
        module_version = ''
    try:
        exec('import ' + module)
        if module_version <> '':
            exec(module + '.require("' + module_version + '")')
        if modules_verbose:
            print 'Checking for module ' + module + module_version + '...OK'
    except:
        if module_optional:
            print 'Checking for module ' + module + module_version + '...SKIPPED'
            modules_skipped.append(module)
        else:
            print 'Checking for module ' + module + module_version + '...FAILED'
            modules_missing.append(module)
if len(modules_skipped):
    print 'Note: The following optional modules are missing:'
    print ' '.join(modules_skipped)
if len(modules_missing):
    print 'Error: The following required modules are missing:'
    print ' '.join(modules_missing)
    sys.exit(1)

app_name = 'search4files'
app_version = '0.2a'

class search4files:
    def __init__(self):
        """Create the main window."""

        # Guess default fileman
        if os.environ.get('DESKTOP_SESSION', os.environ.get('GDMSESSION', ''))[:4] == 'xfce':
            default_fileman = 'Thunar'
            self.open_wrapper = 'exo-open'
        else:
            default_fileman = 'nautilus'
            # Guess suitable file open wrapper
            for path in os.environ.get('PATH', '/usr/bin').split(os.pathsep):
                for wrapper in ['gnome-open', 'exo-open', 'xdg-open']:
                    if os.path.exists(os.path.join(path, wrapper)):
                        self.open_wrapper = wrapper
                        break

        # Parse command line options
        parser = optparse.OptionParser(usage='usage: ' + app_name + ' [options] keywords',
            version=app_name + ' v' + app_version)
        parser.add_option('', '--large-icons', action='store_true', dest='icons_large',
            help='Use large icons')
        parser.add_option('', '--thumbnails', action='store_true', dest='thumbnails',
            help='Use thumbnails')
        #parser.add_option('', '--vertical-mode', action='store_true', dest='vertical',
        #    help='Use vertical layout')
        parser.add_option('', '--iso-time', action='store_true', dest='time_iso',
            help='Display time in iso format')
        parser.add_option('', '--limit', type='int', metavar='LIMIT', dest='limit_results',
            help='Limit number of results')
        parser.add_option('', '--path', help='Search in folder PATH')
        parser.add_option('', '--fileman', help='Use FILEMAN as filemanager')
        parser.add_option('', '--wrapper', metavar='WRAPPER', dest='open_wrapper',
            help='Use WRAPPER to open files')
        parser.add_option('', '--method', dest='method', help='Use METHOD to search')
        parser.add_option('', '--exact', action='store_true', dest='exact', help='Perform exact match')
        parser.add_option('', '--hidden', action='store_true', dest='hidden', help='Include hidden files')
        parser.set_defaults(icons_large=0, thumbnails=0, time_iso=0, limit_results=0,
            path=os.path.expanduser('~'), fileman=default_fileman, exact=0, hidden=0)
        self.options, args = parser.parse_args()
        keywords = ' '.join(args)

        # Guess location of glade file
        glade_file = app_name + '.glade'
        glade_path = os.getcwd()
        if not os.path.exists(os.path.join(glade_path, glade_file)):
            glade_path = os.path.dirname(sys.argv[0])
            if not os.path.exists(os.path.join(glade_path, glade_file)):
                for path in os.environ.get('XDG_DATA_DIRS').split(os.pathsep):
                    if os.path.exists(os.path.join(path, glade_file)):
                        glade_path = path
                        break

        # Load interface from glade file
        try:
            self.xml = gtk.glade.XML(os.path.join(glade_path, glade_file))
            self.xml.signal_autoconnect(self)
        except:
            print 'Error:'
            print 'The glade file could not be found.'
            sys.exit()

        # Alternate layout
        if 0:#if self.options.vertical:
            # move findbar to the top <-- doesn't work
            table = self.xml.get_widget('table_layout')
            findbar = self.xml.get_widget('scrolled_findbar')
            files = self.xml.get_widget('scrolled_files')
            table.remove(findbar); table.remove(files)
            table.attach(findbar, 0, 1, 0, 1, -1, gtk.FILL)
            table.attach(files, 0, 1, 1, 2)
            # change vbox to hbox
            viewport = self.xml.get_widget('viewport_findbar')
            vbox = self.xml.get_widget('vbox_findbar')
            hbox = gtk.HBox()
            for child in vbox.get_children():
                child.reparent(hbox)
            viewport.remove(vbox)
            viewport.add(hbox)
            hbox.show()

        # Prepare significant widgets
        self.window_search = self.xml.get_widget('window_search')
        self.entry_find_text = self.xml.get_widget('entry_find_text')
        self.checkbox_find_exact = self.xml.get_widget('checkbox_find_exact')
        self.checkbox_find_hidden = self.xml.get_widget('checkbox_find_hidden')
        self.checkbox_find_limit = self.xml.get_widget('checkbox_find_limit')
        self.spin_find_limit = self.xml.get_widget('spin_find_limit')
        self.toolbar_find_type = self.xml.get_widget('toolbar_find_type')
        self.combobox_find_method = self.xml.get_widget('combobox_find_method')
        self.button_find_folder = self.xml.get_widget('button_find_folder')
        self.button_find = self.xml.get_widget('button_find')
        self.scrolled_findbar = self.xml.get_widget('scrolled_findbar')
        self.treeview_files = self.xml.get_widget('treeview_files')
        self.menu_file = self.xml.get_widget('menu_file')
        self.statusbar = self.xml.get_widget('statusbar')

        # Set some initial values
        self.icon_cache = {}
        self.icon_theme = gtk.icon_theme_get_default()
        self.checkbox_find_exact.set_active(self.options.exact)
        self.checkbox_find_hidden.set_active(self.options.hidden)
        if self.options.limit_results:
            self.checkbox_find_limit.set_active(1)
            self.checkbox_find_limit.toggled()
            self.spin_find_limit.set_value(self.options.limit_results)
        self.folder_thumbnails = os.path.expanduser('~/.thumbnails/normal/')
        self.button_find_folder.set_filename(self.options.path)

        # Set up keywords completion
        completion = gtk.EntryCompletion()
        self.entry_find_text.set_completion(completion)
        listmodel = gtk.ListStore(gobject.TYPE_STRING)
        completion.set_model(listmodel)
        completion.set_text_column(0)

        # Retrieve available search methods
        methods = (
            ('find', 'find'),
            ('locate', 'locate'),
            ('slocate', 'slocate'),
            ('tracker', 'tracker-search'),
            ('doodle', 'doodle'),
            ('beagle', 'beagle-query')
            )
        bin_dirs = os.environ.get('PATH', '/usr/bin').split(os.pathsep)
        listmodel = gtk.ListStore(gtk.gdk.Pixbuf, gobject.TYPE_STRING)
        method_default = 0
        for method_name, method_binary in methods:
            for path in bin_dirs:
                if os.path.exists(os.path.join(path, method_binary)):
                    if not os.path.islink(os.path.join(path, method_binary)):
                        listmodel.append([self.get_icon_pixbuf('system-search', 16), method_name])
                        if self.options.method == method_name:
                            method_default = len(listmodel) - 1
                        break

        # Prepare method combobox
        cell = gtk.CellRendererPixbuf()
        cell.set_property('xalign', 0)
        self.combobox_find_method.pack_start(cell, True)
        self.combobox_find_method.add_attribute(cell, 'pixbuf', 0)
        cell = gtk.CellRendererText()
        self.combobox_find_method.pack_start(cell, True)
        self.combobox_find_method.add_attribute(cell, 'text', 1)
        self.combobox_find_method.set_model(listmodel)
        self.combobox_find_method.set_active(method_default)

        # Prepare type toolbar
        tooltips = gtk.Tooltips()
        for icon, label, filter_value in (
            ('text-x-generic', 'Documents', 'text'),
            ('image-x-generic', 'Images', 'image'),
            ('audio-x-generic', 'Music', 'audio'),
            ('video-x-generic', 'Videos', 'video')
            ):
            self.toolbar_find_type.insert(self.new_toggle_button(icon, label, filter_value), -1)
        self.toolbar_find_type.show_all()

        # Initialize treeview
        self.treeview_files.append_column(gtk.TreeViewColumn('Preview',
            gtk.CellRendererPixbuf(), pixbuf=0))
        self.treeview_files.append_column(self.new_column('Filename', 1))
        if not self.options.icons_large and not self.options.thumbnails:
            self.treeview_files.append_column(self.new_column('Size', 2, 'filesize'))
            self.treeview_files.append_column(self.new_column('Last modified', 3))
            self.treeview_files.append_column(self.new_column('Location', 4))

        self.entry_find_text.set_text(keywords)
        self.button_find.activate()

# -- helper functions --

    def new_toggle_button(self, icon, label, filter_value):
        togglebutton = gtk.ToggleToolButton()
        togglebutton.set_label(label)
        icon_widget = gtk.image_new_from_icon_name(icon, gtk.ICON_SIZE_MENU)
        togglebutton.set_icon_widget(icon_widget)
        togglebutton.filter_value = filter_value
        return togglebutton

    def new_column(self, label, id, special=None):
        cell_renderer = gtk.CellRendererText()
        column = gtk.TreeViewColumn(label, cell_renderer, text=id)
        if special == 'filesize':
            column.set_cell_data_func(cell_renderer, self.cell_data_func_filesize, id)
        column.set_sort_column_id(id)
        column.set_resizable(1)
        return column

    def cell_data_func_filesize(self, column, cell_renderer, tree_model, tree_iter, id):
        filesize = self.format_size(int(tree_model.get_value(tree_iter, id)))
        cell_renderer.set_property('text', filesize)
        return

    def format_size(self, size):
        if size > 2 ** 30:
            return '%s GB' % (size / 2 ** 30)
        elif size > 2 ** 20:
            return '%s MB' % (size / 2 ** 20)
        elif size > 2 ** 10:
            return '%s kB' % (size / 2 ** 10)
        elif size > -1:
            return '%s B' % size
        else:
            return ''

    def get_selected_filename(self):
        model = self.treeview_files.get_model()
        path = self.treeview_files.get_cursor()[0]
        if not self.options.icons_large and not self.options.thumbnails:
            return model.get_value(model.get_iter(path), 4), model.get_value(model.get_iter(path), 1)
        return model.get_value(model.get_iter(path), 4), model.get_value(model.get_iter(path), 3)

    def open_file(self, filename):
        """Open the file with its default app or the file manager"""

        if stat.S_ISDIR(os.stat(filename).st_mode):
            command = '%s "%s"' % (self.options.fileman, filename)
        else:
            command = '%s "%s"' % (self.open_wrapper, filename)
        try:
            subprocess.Popen(command, shell=True)
        except:
            print 'Error: Could not open ' + filename + '.'
            print 'Note:  The wrapper was ' + self.open_wrapper + '.'

    def get_selected_find_method(self):
        model = self.combobox_find_method.get_model()
        tree_iter = self.combobox_find_method.get_active_iter()
        return model.get_value(tree_iter, 1)

    def get_find_command(self, method, keywords, folder, exact, hidden, limit):
        if method == 'slocate':
            method = 'locate'
        methods = {
            'find': (method + ' ' + folder + ' -ignore_readdir_race -noleaf',
                '-wholename', '-iwholename', '', 0, 0, 0),
            'locate': (method, '', '-i', '-n %i' % limit, 1, 0, 0),
            'tracker': ('tracker-search', '', '', '-l %i' % limit, 1, 1, 0),
            'doodle': (method, '', '-i', '', 0, 0, 1),
            'beagle': ('beagle-query', '', '', '--max-hits=%i' % limit, 1, 1, 0)
            }
        (default, case, nocase, limit_results, path_manual,
            exact_manual, errors_ignore) = methods[method]

        command = default
        if exact:
            command += ' ' + case
        else:
            command += ' ' + nocase
        if limit > 0:
            command += ' ' + limit_results
        command += ' "*%s*"' % keywords
        return command, path_manual, exact_manual, errors_ignore

    def string_wild_match(self, string, keyword, exact):
        return fnmatch.fnmatch(string.lower(), '*' + keyword.lower() + '*')

    def get_mime_type(self, filename):
        mime = xdg.Mime.get_type(filename)
        return mime.media, mime.subtype

    def file_type_is_wanted(self, type):
        wanted = 0
        not_selected = 0
        for button in self.toolbar_find_type.get_children():
            if button.get_active():
                if button.filter_value == type:
                    wanted = 1
            else:
                not_selected += 1
        if wanted or not_selected == self.toolbar_find_type.get_n_items():
            return True
        return False

    def find(self):
        """Do the actual search."""

        self.button_find.set_label('gtk-cancel')
        self.window_search.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
        self.window_search.set_title('Searching for "%s"' % self.entry_find_text.get_text())
        self.statusbar.push(self.statusbar.get_context_id('results'), 'Searching...')
        while gtk.events_pending(): gtk.main_iteration()

        # Reset treeview
        listmodel = gtk.ListStore(gtk.gdk.Pixbuf, gobject.TYPE_STRING,
            gobject.TYPE_INT, gobject.TYPE_STRING, gobject.TYPE_STRING)
        self.treeview_files.set_model(listmodel)
        self.treeview_files.columns_autosize()

        # Retrieve search parameters
        keywords = self.entry_find_text.get_text().replace(' ', '*')
        method = self.get_selected_find_method()
        folder = self.button_find_folder.get_filename()
        exact = self.checkbox_find_exact.get_active()
        hidden = self.checkbox_find_hidden.get_active()
        if self.checkbox_find_limit.get_active():
            limit = self.spin_find_limit.get_value()
        else:
            limit = -1

        if keywords != '':
            # Generate search command
            command, path_manual, exact_manual, error_ignore = self.get_find_command(
                method, keywords, folder, exact, hidden, limit)

            # Set display options
            if not self.options.icons_large and not self.options.thumbnails:
                icon_size = gtk.ICON_SIZE_SMALL_TOOLBAR
            else:
                icon_size = gtk.ICON_SIZE_DIALOG
            if not self.options.time_iso:
                time_format = '%x %X'
            else:
                time_format = '%Y-%m-%w %H:%M %Z'

            # Run search command and capture the results
            search_process = subprocess.Popen(command, stdout=subprocess.PIPE, bufsize=1, shell=True)
            for filename in search_process.stdout:
                while gtk.events_pending(): gtk.main_iteration()
                if self.abort_find or len(listmodel) == limit:
                    break
                filename = filename.split(os.linesep)[0]
                if filename[:7] == 'file://':
                    filename = filename[7:]
                path, name = os.path.split(filename)
                try:
                    if path_manual and not self.string_wild_match(name, keywords, exact):
                        continue
                    else:
                        path_valid = 1
                    if not hidden and self.file_is_hidden(filename):
                        continue
                    if exact_manual and not self.string_wild_match(name, keywords, exact):
                        continue
                    if not self.file_type_is_wanted(self.get_mime_type(filename)[0]):
                        continue
                    if self.options.thumbnails:
                        icon = self.get_thumbnail(filename, icon_size)
                    else:
                        icon = self.get_file_icon(filename, icon_size)
                    filestat = os.stat(filename)
                    size = filestat.st_size
                    modified = time.strftime(time_format, time.gmtime(filestat.st_mtime))
                    if not self.options.icons_large and not self.options.thumbnails:
                        listmodel.append([icon, name, size, modified, path])
                    else:
                        listmodel.append([icon, '%s (%s) %s%s%s%s'
                        % (name, size, os.linesep, modified, os.linesep, path), -1, name, path])
                except:
                    pass # ignore inaccessible and outdated files
                self.treeview_files.set_model(listmodel)
                yield True
            if len(listmodel) == 0:
                if not error_ignore and search_process.poll() <> 0:
                    message = 'Fatal error, search was aborted.'
                else:
                    message = 'No files were found.'
                listmodel.append([None, message, -1, '', ''])
                status = 'No files found for "%s".' % keywords
            else:
                status = '%s files found for "%s".' % (len(listmodel), keywords)
            self.statusbar.push(self.statusbar.get_context_id('results'), status)
        self.treeview_files.set_model(listmodel)

        self.window_search.window.set_cursor(None)
        self.button_find.set_label('gtk-find')
        yield False

    def file_is_hidden(self, filename):
        """Determine if a file is hidden or in a hidden folder"""

        if filename == '': return 0
        path, name = os.path.split(filename)
        if name[0] == '.':
            return 1
        for folder in path.split(os.path.sep):
            if len(folder):
                if folder[0] == '.':
                    return 1
        return 0

    def get_icon_pixbuf(self, name, icon_size=gtk.ICON_SIZE_MENU):
        try:
            return self.icon_cache[name]
        except KeyError:
            icon_size = gtk.icon_size_lookup(icon_size)[0]
            icon = self.icon_theme.load_icon(name, icon_size, 0)
            self.icon_cache[name] = icon
            return icon

    def get_thumbnail(self, path, icon_size=0, use_mime=1):
        """Try to fetch a small thumbnail."""

        filename = '%s%s.png' % (self.folder_thumbnails, md5.new('file://' + path).hexdigest())
        try:
            return gtk.gdk.pixbuf_new_from_file(filename)
        except:
            return self.get_file_icon(path, icon_size, use_mime)

    def get_file_icon(self, path, icon_size=0 ,use_mime=1):
        """Retrieve the file icon."""

        try:
            is_folder = stat.S_ISDIR(os.stat(path).st_mode)
        except:
            is_folder = 0
        if is_folder:
            icon_name='folder'
        else:
            if use_mime:
                try:
                    # Get icon from mimetype
                    media, subtype = self.get_mime_type(path)
                    icon_name = 'gnome-mime-%s-%s' % (media, subtype)
                    return self.get_icon_pixbuf(icon_name, icon_size)
                except:
                    try:
                        # Then try generic icon
                        icon_name = '%s-x-generic' % media
                        return self.get_icon_pixbuf(icon_name, icon_size)
                    except:
                        # Use default icon
                        icon_name = 'gnome-fs-regular'
            else:
                icon_name = 'gnome-fs-regular'
        return self.get_icon_pixbuf(icon_name, icon_size)

# -- events --

    def on_window_search_destroy(self, window):
        gtk.main_quit()

    def on_button_close_clicked(self, button):
        self.window_search.destroy()

    def on_checkbox_find_limit_toggled(self, checkbox):
        self.spin_find_limit.set_sensitive(checkbox.get_active())

    def on_combobox_find_method_changed(self, combobox):
        self.button_find_folder.set_sensitive(self.get_selected_find_method() == 'find')

    def on_button_find_clicked(self, button):
        """Initiate the search thread."""

        # add search term to the completion list
        keywords = self.entry_find_text.get_text()
        completion = self.entry_find_text.get_completion()
        listmodel = completion.get_model()
        listmodel.append([keywords])

        if self.button_find.get_label() == 'gtk-find' and keywords <> '':
            self.abort_find = 0
            task = self.find()
            gobject.idle_add(task.next)
        else:
            self.abort_find = 1

    def on_treeview_files_row_activated(self, treeview, path, view_column):
        folder, filename = self.get_selected_filename()
        filename = os.path.join(folder, filename)
        self.open_file(filename)

    def on_treeview_files_button_pressed(self, treeview, event):        
        if event.button == 3:
            self.menu_file.popup(None, None, None, event.button, event.time)

    def on_menu_open_activate(self, menu):
        folder, filename = self.get_selected_filename()
        self.open_file(os.path.join(folder, filename))

    def on_menu_goto_activate(self, menu):
        folder, filename = self.get_selected_filename()
        self.open_file(folder)

    def on_menu_copy_activate(self, menu):
        folder, filename = self.get_selected_filename()
        clipboard = gtk.clipboard_get('CLIPBOARD')
        clipboard.set_text(os.path.join(folder, filename), -1)

search4files()
gtk.main()